#include <FileSocket.h>
#include <TestSocket.h>
#include <UDT2Socket.h>
#include <RTCPServerSocket.h>
#include <RTCPClientSocket.h>
#include <VRServerSocket.h>
#include <VRClientSocket.h>
#include <RTPThread.h>
#include <Options.h>
#include <Utils.h>
#include <stdio.h>
#include <deque>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <signal.h>
#include <K5.h>
#include <assert.h>


// Global variables.
string version("1.0.H");
string date("6 October 2004");

// Function prototypes.
void 
client(const Options& o);
void 
server(const Options& o);
void
qbss_server(const Options& o);
void
watchdog(const Options& o, const pid_t& pid);
void
k5(const Options& o);
void
welcome();
void
sigterm(int signo);

#if 0
int main(int argc, char **argv)
{
  TCPSocket source;
  string remote_ip("140.173.174.211");
  string local_ip("0.0.0.0");
  source.bind(local_ip, 2630);
  source.listen();	
  int sd=source.accept();
  TCPSocket read_sock(sd, 1500);
  UDPSocket write_sock(1500);
  write_sock.connect(remote_ip, 2630);
  u_int8_t buf[1500];
  while (true) {
    int rcv=read_sock.recv(buf, 1500);	
    if (rcv!=1500)
      cerr << "Short read: " << rcv << endl;
    if (rcv<0)
      break;
    int wrt=write_sock.send(buf, rcv);	
    if (wrt!=rcv)	
      cerr << "Short write: " << rcv << endl;
  }
} 
#endif

int main(int argc, char** argv)
{
  // Print welcome message.
  welcome();
  // Create options object. This automatically parses the commandline options.
  Options o(argc, argv);
  // Internal debugging.
  string debug_log = "/tmp/vtp-debug.log";
  string info_log = "/tmp/vtp-info.log";
  string warning_log = "/tmp/vtp-wraning.log";
  string error_log = "/tmp/vtp-error.log";
  // Set debug level.
  switch (o._debug_level) {
  case 3:
    Logsystem::debug_logger()->enable_logging();
    Logsystem::debug_logger()->set_stream(debug_log);
  case 2:
    Logsystem::info_logger()->enable_logging();
    Logsystem::info_logger()->set_stream(info_log);
  case 1:
    Logsystem::warning_logger()->enable_logging();
    Logsystem::warning_logger()->set_stream(warning_log);
  case 0:
    Logsystem::error_logger()->enable_logging();
    Logsystem::error_logger()->set_stream(error_log);
  default:
    break;
  }	 
  // External debugging.
  // Open system log.
  openlog("vsocket", LOG_NDELAY | LOG_PERROR, LOG_USER);
  // openlog("vsocket", LOG_NDELAY, LOG_USER);
  // Set debug level.
  switch (o._debug_level) {
  case 3:
    setlogmask(LOG_UPTO(LOG_DEBUG));
    break;
  case 2:
    setlogmask(LOG_UPTO(LOG_INFO));
    break;
  case 1:
    setlogmask(LOG_UPTO(LOG_WARNING));
    break;
  case 0:
    setlogmask(LOG_UPTO(LOG_ERR));
    break;
  default:
    break;
  }
  
  // Determine mode and launch appropriate processes.
  pid_t pid=0;
  int ret=0;
  try {
    switch (o._mode) {
    case K5T:
      k5(o);
      break;
    case FILE2NET:
    case IN2NET:
    case TVG2NET:
#ifdef FORK
      pid=fork();
#else
      pid=0;
#endif
      switch (pid) {
      case 0:
	// Child process runs server().
	server(o);
	// qbss_server(o);					 
	// Child process completed, kill parent process.
	cout << "Server process complete. Killing watchdog process\n";
	pid=getppid();
	ret=kill(pid, SIGTERM);
	if (ret!=0)
	  syslog(LOG_ERR, "Kill error: %s", strerror(errno));
	else
	  cout << "Watchdog process killed...\n";
	_exit(0);
	break;
      case -1:
	// Fork failure.
	syslog(LOG_ERR, "fork()-ing error: %s\n", strerror(errno));
	exit(1);
	break;
      default:
	// Parent process runs watchdog(). This process terminates
	// child process if maximum time limit exceeded.
	watchdog(o, pid);
	break;
      }
      break;
    case NET2FILE:
    case NET2OUT:
    case NET2TVR:
#ifdef FORK
      pid=fork();
#else
      pid=0;
#endif
      switch (pid) {
      case 0:
	// Child process runs client().
	client(o);
	// Child process completed, kill parent process.
	cout << "Client process complete. Killing watchdog process\n";
	pid=getppid();
	ret=kill(pid, SIGTERM);
	if (ret!=0)
	  syslog(LOG_ERR, "Kill error: %s", strerror(errno));
	else
	  cout << "Watchdog process killed...\n";
	_exit(0);
	break;
      case -1:
	// Fork failure.
	syslog(LOG_ERR, "fork()-ing error: %s\n", strerror(errno));
	exit(1);
	break;
      default:
	// Parent process runs watchdog(). This process terminates
	// child process if maximum time limit exceeded.
	watchdog(o, pid);
	break;
      }
      break;
    default:
      syslog(LOG_ERR, "Uknown test mode. There was an error " 
	     "in the command line. Please try again.");
      break;
    } 
  } catch (exception e) {
    syslog(LOG_ERR, "Exception: %s", e.what());
  }
  closelog();
  return(0);
}

void
qbss_server(const Options& o)
{
	 cout << "Qbss server was configured with the following parameters:\n";
	 o.dump(cout);
	 const int& mtu=o._t_spec._mtu;
	 UDPSocket out_sock(mtu);
// out_sock.set_ip_tos(IPTOS_QBSS);
	 int ret=out_sock.connect(o._dst_addr, o._data_port);
	 assert(ret==0);
	 u_int8_t *buf=new u_int8_t[mtu];
	 Timer t;
	 t.start();
	 for (int i=0; i<mtu; ++i)
		  buf[i]='0'+(i%10);
	 int j=0;
	 while (true) {
		  out_sock.send(buf, mtu);
		  for (int k=0; k<50000; k++)
				;
		  t.stop();
		  if (t.elapsed()>o._tvg_duration)
				break;
	 }
	 delete [] buf;
}

void 
client(const Options& o)
{
  cout << "Client was configured with the following parameters:\n";
  o.dump(cout);
  string local_ip("0.0.0.0");
  const int& mtu=o._t_spec._mtu;

  RTPSession rtp_session(o._t_spec);	 

  // Outgoing.
  Socket* out_sock=0;
  // RTCP.
  Socket* rtcp_accept_sock=0;
  Socket* rtcp_transport_sock=0;
  RTCPSocket* rtcp_sock=0;
  // RTP.
  Socket* rtp_accept_sock=0;
  Socket* rtp_transport_sock=0;
  Socket* rtp_sock=0;
  
  switch (o._mode) {
  case NET2FILE:
    {
      // 1. Outgoing.
      // TODO: rename remote_ip parameter to remote_address.
      out_sock=new FileSocket(mtu);
      assert( out_sock->connect(o._dst_addr, 0)==0 );
      
      // 2. RTCP.
      rtcp_accept_sock=new TCPSocket(mtu);
      rtcp_accept_sock->set_so_reuseaddr(1);
      rtcp_accept_sock->bind(local_ip, o._control_port);
      rtcp_accept_sock->listen();
      int sd=rtcp_accept_sock->accept();
      assert(sd>0);
      rtcp_transport_sock=new TCPSocket(sd, mtu);
      rtcp_sock=new RTCPClientSocket(*rtcp_transport_sock, 
				     rtp_session,
				     mtu);
      
      // 3. RTP.
      rtp_accept_sock=new TCPSocket(mtu);
      rtp_accept_sock->set_so_reuseaddr(1);
      rtp_accept_sock->bind(local_ip, o._data_port);
      rtp_accept_sock->listen();
      sd=rtp_accept_sock->accept();
      assert(sd>0);
      rtp_transport_sock=new TCPSocket(sd, mtu);
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,
			     mtu);		  
    }
    break;
  case NET2OUT:
    {
      // 1. Outgoing.
      // TODO: rename remote_ip parameter to remote_address.
      out_sock=new TCPSocket(mtu);
      assert( out_sock->connect(o._dst_addr, 2630)==0 );
      
      // 2. RTCP.
      rtcp_accept_sock=new TCPSocket(mtu);
      rtcp_accept_sock->set_so_reuseaddr(1);
      rtcp_accept_sock->bind(local_ip, o._control_port);
      rtcp_accept_sock->listen();
      int sd=rtcp_accept_sock->accept();
      assert(sd>0);
      rtcp_transport_sock=new TCPSocket(sd, mtu);
      rtcp_sock=new RTCPClientSocket(*rtcp_transport_sock, 
				     rtp_session,
				     mtu);
      
      // 3. RTP.
      rtp_accept_sock=new TCPSocket(mtu);
      rtp_accept_sock->set_so_reuseaddr(1);
      rtp_accept_sock->bind(local_ip, o._data_port);
      rtp_accept_sock->listen();
      sd=rtp_accept_sock->accept();
      assert(sd>0);
      rtp_transport_sock=new TCPSocket(sd, mtu);
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,
			     mtu);

    }
    break;
  case NET2TVR:
    {
      // 1. Outgoing.
      // TODO: rename remote_ip parameter to remote_address.
      out_sock=new TestSocket(mtu);
      assert( out_sock->connect(o._dst_addr, 0)==0 );
		  
      // 2. RTCP.
      rtcp_accept_sock=new TCPSocket(mtu);
      rtcp_accept_sock->set_so_reuseaddr(1);
      rtcp_accept_sock->bind(local_ip, o._control_port);
      rtcp_accept_sock->listen();
      int sd=rtcp_accept_sock->accept();
      assert(sd>0);
      rtcp_transport_sock=new TCPSocket(sd, mtu);
      rtcp_sock=new RTCPClientSocket(*rtcp_transport_sock, 
				     rtp_session,
				     mtu);
		  
      // 3. RTP.
      rtp_accept_sock=new TCPSocket(mtu);
      rtp_accept_sock->set_so_reuseaddr(1);
      rtp_accept_sock->bind(local_ip, o._data_port);
      rtp_accept_sock->listen();
      sd=rtp_accept_sock->accept();
      assert(sd>0);
      rtp_transport_sock=new TCPSocket(sd, mtu);
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,
			     mtu);
    }
    break;
  default:
    {
      syslog(LOG_ERR, "client(): unknown mode\n");
    }
    break;
  }

  // 4. VRClientSocket.
  VRClientSocket client(*rtp_sock, *out_sock, *rtcp_sock,
			o._tvg_duration,
			o._t_spec._mtu,
			o._rtcp_interval);
  client.rtcp_proc();
  client.rtp_proc();

  delete out_sock;
  delete rtcp_accept_sock;
  delete rtcp_transport_sock;
  delete rtcp_sock;
  delete rtp_accept_sock;
  delete rtp_transport_sock;
  delete rtp_sock;
}

void server(const Options& o)
{
  cout << "Server was configured with the following parameters:\n";
  o.dump(cout);
  const int& mtu=o._t_spec._mtu;
  
  RTPSession rtp_session(o._t_spec);	 
  
  // Outgoing.
  Socket* in_accept_sock=0;
  Socket* in_sock=0;
  // RTCP.
  Socket* rtcp_transport_sock=0;
  RTCPSocket* rtcp_sock=0;
  // RTP.
  Socket* rtp_transport_sock=0;
  Socket* rtp_sock=0;
  
  switch (o._mode) {
  case FILE2NET:
    {
      // 1. RTCP.
      rtcp_transport_sock=new TCPSocket(mtu);
      fprintf(stderr, "RTCP Socket Connecting..\n");
      assert( rtcp_transport_sock->connect(o._dst_addr, 
					   o._control_port) == 0 );
      rtcp_sock=new RTCPServerSocket(*rtcp_transport_sock, 
				     rtp_session, mtu);
	   
      // Wait for client to setup RTP accept socket.
      usleep(1000000);

      // 2. RTP.
      rtp_transport_sock=new TCPSocket(mtu);
      fprintf(stderr, "RTP Socket Connecting..\n");
      assert( rtp_transport_sock->connect(o._dst_addr,
					  o._data_port) == 0 );
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,  mtu); 

      // 3. Incoming.
      in_accept_sock=new FileSocket(mtu);
      assert( in_accept_sock->bind(o._src_addr, 0)==0 );
      fprintf(stderr, "FILE Socket Listening for connection...\n");
      assert( in_accept_sock->listen()==0 );
      int sd=in_accept_sock->accept();
      assert( sd>0 );
      in_sock=new FileSocket(sd, mtu);
    }
    break;
  case IN2NET:
    {
      // 1. RTCP.
      rtcp_transport_sock=new TCPSocket(mtu);
      assert( rtcp_transport_sock->connect(o._dst_addr, 
					   o._control_port) == 0 );
      rtcp_sock=new RTCPServerSocket(*rtcp_transport_sock, 
				     rtp_session,
				     mtu);

      // Wait for client to setup RTP accept socket.
      usleep(1000000);

      // 2. RTP.
      rtp_transport_sock=new TCPSocket(mtu);
      assert( rtp_transport_sock->connect(o._dst_addr,
					  o._data_port) == 0 );
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,
			     mtu); 

      // 3. Incoming.
      in_accept_sock=new TCPSocket(mtu);
      assert( in_accept_sock->bind(o._src_addr, 2630)==0 );
      assert( in_accept_sock->listen()==0 );
      int sd=in_accept_sock->accept();
      assert( sd>0 );
      in_sock=new TCPSocket(sd, mtu);
    }
    break;
  case TVG2NET:
    {
      // 1. RTCP.
      rtcp_transport_sock=new TCPSocket(mtu);
      assert( rtcp_transport_sock->connect(o._dst_addr, 
					   o._control_port) == 0 );
      rtcp_sock=new RTCPServerSocket(*rtcp_transport_sock, 
				     rtp_session,
				     mtu);

      // Wait for client to setup RTP accept socket.
      usleep(1000000);

      // 2. RTP.
#ifdef UDT_IMPL
      rtp_transport_sock=new UDT2Socket(mtu);
      assert( rtp_transport_sock->connect(o._dst_addr,
					  o._data_port) == 0 );
#else
      rtp_transport_sock=new TCPSocket(mtu);
      assert( rtp_transport_sock->connect(o._dst_addr,
					  o._data_port) == 0 );
#endif
      rtp_sock=new RTPSocket(*rtp_transport_sock,
			     rtp_session,
			     mtu); 

      // 3. Incoming.
      in_accept_sock=new TestSocket(mtu);
      assert( in_accept_sock->bind(o._src_addr, 0)==0 );
      assert( in_accept_sock->listen()==0 );
      int sd=in_accept_sock->accept();
      assert( sd>0 );
      in_sock=new TestSocket(mtu);
    }
    break;
  default:
    {
      syslog(LOG_ERR, "client(): unknown mode\n");
    }
    break;
  }

  // 4. VRServerSocket.
  VRServerSocket server(*in_sock, *rtp_sock, *rtcp_sock,
			o._tvg_duration,
			o._t_spec._mtu,
			o._rtcp_interval);
  server.rtcp_proc();
  server.rtp_proc();

  delete in_accept_sock;
  delete in_sock;
  delete rtcp_transport_sock;
  delete rtcp_sock;
  delete rtp_transport_sock;
  delete rtp_sock;
}

void
k5(const Options& o)
{
  cout << "K5 was configured with the following parameters:\n";
  o.dump(cout);
  cout << endl;
  K5 k(o._k5udp, o._mjd, o._dst_addr, o._data_port);
  k.run();
}

void 
watchdog(const Options &o, const pid_t& pid) 
{
	 int ret=o._max_time;
	 if (ret==-1) {
		  // No watchdog required.
		  cout << "Watchdoq not required. Exiting process...\n";
		  exit(0);
	 }

	 cout << "Launching watchdog process...\n" 
			<< "Watchdog will terminate child transfer process in " 
			<< o._max_time << " s\n"
			<< "Watchdog going to sleep...\n";
	 // Setup signal handler to allow exiting client to signal watchdog
	 // to exit.
	 if (signal(SIGTERM, sigterm)==SIG_ERR) {
		  cerr << "Watchdog unable to install signal handler. Exiting...\n";
		  exit(1);
	 }
	 while (ret>0) {
		  ret=sleep(o._max_time);
	 }
	 cout << "Watchdog process awake...\n"
			<< "Child transport process " << pid 
			<< " has exceeded maximum allowed transfer time.\n"
			<< "Watchdog process terminating child process " << pid << endl;
	 ret=kill(pid, SIGKILL);
	 cout << "Child transport process killed...\n";
	 if (ret!=0) {
		  syslog(LOG_ERR, "Kill error: %s", strerror(errno));
		  exit(1);
	 }
	 cout << "Exiting watchdog process.\n";
}

void
sigterm(int signo)
{
	 switch (signo) {
	 case SIGTERM:
		  syslog(LOG_ERR, "Watchdog, received terminate signal. Exiting...");
		  exit(0);
		  break;
	 default:
		  syslog(LOG_ERR, "Watchdog, unhandled signal: %d. Exiting...", signo);
		  break;
	 }
	 exit(0);
}

void
welcome()
{
	 cerr << "***********************************"
			<< "**********************************" 
			<< endl
			<< "*                       "
			<< "VLBI Transfer Protocol"
			<< "                      *" 
			<< endl
			<< "***********************************"
			<< "**********************************" 
			<< endl
			<< "Version:                                 " << version << endl
			<< "Date:                                    " << date << endl
			<< "Author:                                  "
			<< "dlapsley@haystack.mit.edu" 
			<< endl << endl
			<< "This software is intended to provide a reference implementation for" 
			<< endl
			<< "the draft VSI-E specification. It provide RTP framing of data and" 
			<< endl
			<< "RTCP control channel messaging. It also provides high performance" 
			<< endl
			<< "rate-based flow control designed to maximize end to end throughput" 
			<< endl
			<< "in a TCP-friendly manner. "
			<< endl << endl
			<< "This software is released subject to the terms of the MIT software" 
			<< endl
			<< "license. For more information please refer to the README file. Use," 
			<< endl
			<< "distribution and re-use of this software are highly encouraged, " 
			<< endl
			<< "subject to the terms of the MIT license." 
			<< endl << endl
			<< "Please send comments suggestions or bug reports to:  " << endl
			<< "    dlapsley@haystack.mit.edu" << endl
			<< "***********************************"
			<< "**********************************"  
			<< endl << endl;
}


	 
