/*
 *  VRSocket.h
 *
 *  Created by David Lapsley on Mon Feb 23 2004.
 *
 */

/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#ifndef VRSOCKET_H
#define VRSOCKET_H

#include <common.h>
#include <Socket.h>
#include <RTCPSocket.h>
#include <Thread.h>
#include <string>
#include <iostream>
#include <syslog.h>

/** VRSocket class
 * This class implements an RTP server socket. It inherits from VRSocketA.
 */
class VRSocket: public Thread
{
  protected:
	 /** Input socket.
	  * Reference to a conneccted input socket.
	  */
  Socket& _in_sock;
	 /** Output socket.
	  * Reference to a connected output socket.
	  */
  Socket& _out_sock;
	 /** RTCP Socket.
	  * Reference to a connected RTCP socket.
	  */
  RTCPSocket& _rtcp_sock;
	 /** TVG Duration.
	  * Duration of TVG interval.
	  */
  double _tvg_duration;
	 /** Network Parameter.
	  * MTU.
	  */
  int _mtu;
	 /** RTCP Parameter.
	  * Interval between RTCP transmissions/receptions.
	  */
  double _rtcp_interval;
	 /** Network.
	  * Buffer for storing data.
	  */
  u_int8_t* _buf;
	 /** Flag.
	  * Indicated whether process is active or not.
	  */
  bool _active;
  public:
	 /** Processing function.
	  * This function is the entry point for RTP processing in this class. 
	  * It runs in the calling mode context.
	  */
  virtual void rtp_proc() = 0;
	 /** Processing function.
	  * Start RTCP thread execution.
	  */
  virtual void rtcp_proc() = 0;
	 /** Thread callback.
	  * This function is called when the base class method "thread_create()" is
	  * called. This results in thread_func() being executed in it's own
	  * thread. Depending on the mode, either recv_proc() or send_proc() will
	  * be called by thread_func(). This means that recv_proc() or send_proc()
	  * will be running in their own thread of execution.
	  */
  virtual void* thread_func() = 0;
	 /** Constructor.
	  * This constructor creates a socket and sets the internal MTU member
	  * to be equal to the supplied parameter.
	  * @param in_sock input socket reference.
	  * @param out_sock output socket reference.
	  * @param rtcp_sock rtcp socket reference.
	  * @param tvg_duration Duration of test vector generation.
	  * @param mtu size of packet payload in bits.
	  * @param rtcp_interval maximum interval between transmission of RTCP packets.
	  */
  VRSocket(Socket& in_sock,
	   Socket& out_sock,
	   RTCPSocket& rtcp_sock,
	   const double& tvg_duration, 
	   const u_int32_t& mtu, 
	   const double& rtcp_interval);
	 /** Copy constructor.
	  * Copies values of other VRSocket into this object.
	  * @param s VRSocket to copy.
	  */
  VRSocket(const VRSocket &s);
	 /** Destructor. */
  virtual ~VRSocket();
};

ostream& operator<<(ostream& os, const VRSocket& s);


#endif

